<?php

    /** @noinspection PhpMissingFieldTypeInspection */

    namespace TamerLib\Objects;

    use Exception;
    use RuntimeException;
    use TamerLib\Classes\Utilities;

    class WorkerConfiguration
    {
        /**
         * @var string
         */
        private $worker_id;

        /**
         * @var string
         */
        private $host;

        /**
         * @var int
         */
        private $port;

        /**
         * @var string|null
         */
        private $password;

        /**
         * @var int|null
         */
        private $database;

        /**
         * @var int|null
         */
        private $channel;

        /**
         * WorkerConfiguration constructor.
         *
         * @param array $data
         * @throws Exception
         */
        public function __construct(array $data=[])
        {
            $this->worker_id = $data['worker_id'] ?? Utilities::generateRandomString();
            $this->host = $data['host'] ?? '127.0.0.1';
            $this->port = $data['port'] ?? null;
            $this->password = $data['password'] ?? null;
            $this->database = $data['database'] ?? 0;
            $this->channel = $data['channel'] ?? 0;
        }

        /**
         * Constructs object from environment variables.
         *
         * @return WorkerConfiguration
         * @throws Exception
         */
        public static function fromEnvironment(): WorkerConfiguration
        {
            if(getenv('TAMER_WORKER') !== 'true')
            {
                throw new RuntimeException('Process is not running as a worker.');
            }

            $data = [
                'worker_id' => getenv('TAMER_WORKER_ID'),
                'host' => getenv('TAMER_WORKER_HOST'),
                'port' => getenv('TAMER_WORKER_PORT'),
                'password' => getenv('TAMER_WORKER_PASSWORD'),
                'database' => getenv('TAMER_WORKER_DATABASE'),
                'channel' => getenv('TAMER_WORKER_CHANNEL'),
            ];

            return new WorkerConfiguration($data);
        }

        /**
         * @return string
         */
        public function getWorkerId(): string
        {
            return $this->worker_id;
        }

        /**
         * @return string
         */
        public function getHost(): string
        {
            return $this->host;
        }

        /**
         * @param string $host
         */
        public function setHost(string $host): void
        {
            $this->host = $host;
        }

        /**
         * @return int
         */
        public function getPort(): int
        {
            return $this->port;
        }

        /**
         * @param int $port
         */
        public function setPort(int $port): void
        {
            $this->port = $port;
        }

        /**
         * @return string|null
         */
        public function getPassword(): ?string
        {
            return $this->password;
        }

        public function setPassword(?string $password): void
        {
            $this->password = $password;
        }

        /**
         * @return int|null
         */
        public function getDatabase(): ?int
        {
            return $this->database;
        }

        /**
         * @param int|null $database
         * @return void
         */
        public function setDatabase(?int $database): void
        {
            $this->database = $database;
        }

        /**
         * @return int
         */
        public function getChannel(): int
        {
            return $this->channel;
        }

        /**
         * @param int $channel
         * @return void
         */
        public function setChannel(int $channel=0): void
        {
            $this->channel = $channel;
        }

        /**
         * Outputs environment variables for worker.
         *
         * @return array
         */
        public function toEnvironment(): array
        {
            return [
                'TAMER_WORKER' => 'true',
                'TAMER_WORKER_ID' => $this->getWorkerId(),
                'TAMER_WORKER_HOST' => $this->getHost(),
                'TAMER_WORKER_PORT' => $this->getPort(),
                'TAMER_WORKER_PASSWORD' => $this->getPassword(),
                'TAMER_WORKER_DATABASE' => $this->getDatabase(),
                'TAMER_WORKER_CHANNEL' => $this->getChannel(),
            ];
        }
    }